# $Id: tserver.rb,v 1.2 2002/11/10 11:16:54 tommy Exp $

require 'socket'

class TServer

  def initialize(*args)
    @min_servers = 5
    @max_servers = 50
    @max_request_per_child = 50
    @children = []
    @connections = []
    if args.length == 1 and args[0].is_a? TCPSocket then
      @sock = args[0]
    else
      @sock = TCPServer::new(*args)
    end
    trap 'SIGCHLD' do on_child_exit end
    trap 'SIGTERM' do
      Process::kill 'SIGTERM', *@children rescue nil
      exit
    end
    trap 'SIGHUP' do
      Process::kill 'SIGTERM', *@children rescue nil
    end
    trap 'SIGUSR1' do
      @to_child.each_value do |f| f.puts 'exit' end
    end
  end

  attr_reader :sock
  attr_accessor :min_servers, :max_servers, :max_request_per_child

  def start(&block)
    @from_child, @to_parent = IO::pipe
    @to_child = {}
    if block == nil then
      raise 'block required'
    end
    @min_servers.times do
      make_child block
    end
    loop do
      if IO::select([@from_child], nil, nil, 1) then
	from_child @from_child.gets.chomp
      end
      if @children.size < @min_servers or (@connections.size >= @children.size and @children.size < @max_servers) then
	make_child block
      end
    end
  end

  private
  def on_child_exit()
    exited_pids = []
    @children.each do |pid|
      begin
	if Process::waitpid(pid, Process::WNOHANG) then
	  exited_pids << pid
	end
      rescue Errno::ECHILD	# for Linux bug?
      end
    end
    exited_pids.each do |pid|
      @children.delete pid
      @connections.delete pid
      @to_child[pid].close
      @to_child.delete pid
    end
  end

  def from_child(str)
    pid, *args = str.split
    case args[0]
    when 'connect'
      @connections << pid.to_i
    when 'disconnect'
      @connections.delete pid.to_i
    end
  end

  def from_parent(str)
    case str
    when 'exit'
      exit
    end
  end

  def make_child(block)
    pipe = IO::pipe
    pid = fork do
      @to_child.each_value do |f| f.close end
      @from_parent = pipe[0]
      pipe[1].close unless pipe[1].closed?
      @from_child.close
      trap 'SIGCHLD', 'SIG_IGN'
      trap 'SIGTERM', 'SIG_DFL'
      trap 'SIGHUP', 'SIG_DFL'
      trap 'SIGUSR1', 'SIG_DFL'
      cnt = 0
      while @max_request_per_child == 0 or cnt < @max_request_per_child
	r, = IO::select([@sock, @from_parent])
	if r.include? @from_parent then
	  from_parent @from_parent.gets.chomp
	  next
	end
	s = @sock.accept
	@to_parent.puts "#{$$} connect"
	block.call(s)
	s.close unless s.closed?
	@to_parent.puts "#{$$} disconnect"
	cnt += 1
      end
    end
    @children << pid
    pipe[0].close
    @to_child[pid] = pipe[1]
  end

end
